# potrebne biblioteke
import subprocess
import threading
import atexit
import tkinter as tk
from tkinter import ttk, Text, PhotoImage


# Provjeri je li ttkthemes dostupan; ako nije, instaliraj ga
try:
    import ttkthemes
except ImportError:
    print("ttkthemes nije pronađen. Instaliram...")
    subprocess.run(["pip", "install", "ttkthemes"])

# učitavanje definiranih opisa problema i sl.
from strings import *

# Global variable for font size
global_font_size = 12

run_button = None
input_button = None

# Global variable for subprocess
process = None

def run_script(script_name):
    global process
    try:
        clear_text(text_output)
        
        # Terminate the existing subprocess if it's running
        if process and process.poll() is None:
            process.terminate()
        
        # Define the completion callback to be called after the subprocess completes
        def on_completion():
            display_saved_image("./images/"+script_name+".png")

        # Pokreni podproces za izvršavanje skripte u zasebnoj niti
        process_thread = threading.Thread(target=run_subprocess, args=(script_name, on_completion))
        process_thread.start()

    except subprocess.CalledProcessError as e:
        text.insert(tk.END, SCRIPT_EXECUTION_ERROR.format(e) + "\n")


def run_subprocess(script_name, completion_callback):
    global process
    try:
        # Pokreni podproces za izvršavanje skripte
        process = subprocess.Popen(["python", "./scripts/"+script_name+".py"], stdout=subprocess.PIPE, text=True, bufsize=1, universal_newlines=True)

        # Omogući tekstualno polje za modificiranje
        text_output.config(state=tk.NORMAL)

        def update_output():
            # Čitaj i prikazuj izlaz podprocesa
            line = process.stdout.readline()
            if line:
                text_output.insert(tk.END, line)
                text_output.see(tk.END)  # Automatski klizni prikaz na kraj polja teksta
                text_output.update_idletasks()  # Ažuriraj polje teksta
                # Pokreni sljedeći poziv update_output nakon kratke pauze
                root.after(10, update_output)
            else:
                # Ako nema više linija, onemogući tekstualno polje nakon završetka skripte
                text_output.config(state=tk.DISABLED)
                text.insert(tk.END, SCRIPT_EXECUTION_COMPLETED + "\n")
                # Call the completion callback when the subprocess completes
                completion_callback()

        # Pokreni prvi poziv update_output
        update_output()

    except subprocess.CalledProcessError as e:
        text.insert(tk.END, SCRIPT_EXECUTION_ERROR.format(e) + "\n")


def show_input(input_text):
    # Prikazi ulazne podatke u novom prozoru
    input_window = tk.Toplevel(root)
    input_window.title("Ulazni podaci")
    input_window.geometry("500x500")


    # Create a Text widget to display input_text
    input_text_widget = Text(input_window, wrap=tk.WORD, width=100, height=100, state=tk.NORMAL, bg=style.lookup("TFrame", "background"), fg="white", font=("Arial", global_font_size))
    input_text_widget.insert(tk.END, input_text)
    input_text_widget.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
    input_text_widget.config(state=tk.DISABLED)

def show_problem(option):
    global run_button, input_button
    # Briši prethodne gumbe ako postoje
    if run_button:
        run_button.destroy()
    if input_button:
        input_button.destroy()

    # Inicijalizacija varijabli za naziv skripte, naziv problema i opis problema
    script_name = ""
    problem_name = ""
    problem_description = ""
    input_text = ""

    match option:
        case 1:
            script_name = "ruksak"
            problem_name = KNAPSACK_TITLE
            problem_description = KNAPSACK_DESCRIPTION
            input_text = KNAPSACK_INPUT
        case 2:
            script_name = "ruksakSaKlasama"
            problem_name = MULTIPLE_CHOICE_KNAPSACK_TITLE
            problem_description = MULTIPLE_CHOICE_KNAPSACK_DESCRIPTION
            # input_text = MULTIPLE_CHOICE_KNAPSACK_INPUT
        case 3:
            script_name = "Sudoku"
            problem_name = SUDOKU_TITLE
            problem_description = SUDOKU_DESCRIPTION
            input_text = SUDOKU_INPUT
        case 4:
            script_name = "rezanje"
            problem_name = CUTTING_STOCK_TITLE
            problem_description = CUTTING_STOCK_DESCRIPTION
            input_text = CUTTING_STOCK_INPUT
        case 5:
            script_name = "CVRP"
            problem_name = CVRP_TITLE
            problem_description = CVRP_DESCRIPTION
            input_text = CVRP_INPUT
        case _:
            print("Nevažeća opcija skripte")
            return

    # Ubaci naziv skripte i opis
    clear_text(text)
    text.config(state=tk.NORMAL)
    text.insert(tk.END, problem_name + "\n\n")
    text.insert(tk.END, problem_description)
    text.update_idletasks()
    text.config(state=tk.DISABLED)

    run_button = ttk.Button(root, text="Pokreni", command=lambda: run_script(script_name), style="White.TButton")
    run_button.pack(side=tk.RIGHT, padx=10)

    input_button = ttk.Button(root, text="Ulazni podaci", command=lambda: show_input(input_text), style="White.TButton")
    input_button.pack(side=tk.RIGHT, padx=10)

def button_func_1():
    show_problem(1)

def button_func_2():
    show_problem(2)

def button_func_3():
    show_problem(3)

def button_func_4():
    show_problem(4)

def button_func_5():
    show_problem(5)

def clear_text(field):
    # Omogući tekstualno polje za modificiranje
    field.config(state=tk.NORMAL)
    # Obriši sav sadržaj
    field.delete(1.0, tk.END)
    # Postavi stanje natrag na onemogućeno
    field.config(state=tk.DISABLED)

def resize_text(event):
    global global_font_size
    current_width = root.winfo_width()
    global_font_size = min(max(int(12 * current_width / 1000), 8), 16)
    text.configure(font=("Arial", global_font_size))
    text_output.configure(font=("Arial", global_font_size))
    style.configure("White.TButton", foreground="white", background=style.lookup("TFrame", "background"), font=("Arial", global_font_size))
    style.configure("BoldWhite.TLabel", foreground="white", background=style.lookup("TFrame", "background"), font=("Arial", int(global_font_size*1.2), "bold"))

def display_saved_image(image_path):
    window = tk.Toplevel(root)
    window.title("Plot promjene rješenja")
    
    image = tk.PhotoImage(file=image_path)
    img_label = tk.Label(window, image = image)
    img_label.grid(row=4, column=0)
    img_label.image = image

	
	
# Kreiraj glavni Tkinter prozor
root = tk.Tk()
root.title("Primjeri rješavanja optimizacijskih problema genetskim algoritmima")

# Zamijeni ikonu na traci s ikonom iz ikona.png
root.iconphoto(True, tk.PhotoImage(file="./images/ikona.png"))

# Primijeni temu "equilux" iz ttkthemes
style = ttkthemes.ThemedStyle(root)
style.set_theme("equilux")

# Kreiraj okvir za bočnu traku
side_bar_frame = ttk.Frame(root, width=150, style="TFrame")
side_bar_frame.pack(side=tk.LEFT, fill=tk.Y)

# Dodaj naslov bočnoj traci s bijelim tekstom, podebljanim i većim fontom
sidebar_title = ttk.Label(side_bar_frame, text=SIDEBAR_TITLE, style="BoldWhite.TLabel")
sidebar_title.pack(pady=10, padx=10)

# Kreiraj gumbe i poveži ih s funkcijama
button_1 = ttk.Button(side_bar_frame, text=BUTTON_1, command=button_func_1, style="White.TButton")
#button_2 = ttk.Button(side_bar_frame, text=BUTTON_2, command=button_func_2, style="White.TButton")
button_3 = ttk.Button(side_bar_frame, text=BUTTON_3, command=button_func_3, style="White.TButton")
button_4 = ttk.Button(side_bar_frame, text=BUTTON_4, command=button_func_4, style="White.TButton")
button_5 = ttk.Button(side_bar_frame, text=BUTTON_5, command=button_func_5, style="White.TButton")

# Smjesti gumbe u bočnu traku
button_1.pack(pady=10, fill=tk.X)
#button_2.pack(pady=10, fill=tk.X)
button_3.pack(pady=10, fill=tk.X)
button_4.pack(pady=10, fill=tk.X)
button_5.pack(pady=10, fill=tk.X)

# Definiraj prilagođene stilove
style.configure("White.TButton", foreground="white", background=style.lookup("TFrame", "background"), font=("Arial", global_font_size))
style.configure("BoldWhite.TLabel", foreground="white", background=style.lookup("TFrame", "background"), font=("Arial", 14, "bold"))
style.configure("TFrame", background=style.lookup("TFrame", "background"))

# Kreiraj polje teksta za glavni dio GUI-a s modernim stilom i pozadinskom bojom
text = Text(root, wrap=tk.WORD, width=50, height=10, state=tk.NORMAL, bg=style.lookup("TFrame", "background"), fg="white", font=("Arial", global_font_size))
text.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

# Postavi prevedeni tekst o genetskim algoritmima u glavnom polju teksta
text.insert(tk.END, GENETIC_ALGORITHMS_TITLE + "\n\n")
text.insert(tk.END, GENETIC_ALGORITHMS_DESCRIPTION)

# Kreiraj zasebno polje teksta za izlaz skripte
text_output = Text(root, wrap=tk.WORD, width=50, height=10, state=tk.DISABLED, bg=style.lookup("TFrame", "background"), fg="white", font=("Arial", global_font_size))
text_output.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)

# Bind the resize_text function to the Configure event of the root window
root.bind("<Configure>", resize_text)

# Pokreni Tkinter petlju događaja
root.mainloop()


# Register a function to terminate subprocess when the program exits
def on_exit():
    global process
    if process and process.poll() is None:
        process.terminate()
atexit.register(on_exit)

