import random
import matplotlib.pyplot as plt
import pygad
import numpy
from pygad.utils.parent_selection import ParentSelection



numberOfMutations = 3
fitness_values = []



def read():
    array = []
    # Open the input file for reading
    with open('./input/input_sudoku.txt', 'r') as file:
        for line in file:
            if line.strip().replace("\n", "").__len__() > 0:
                # Split the line into individual numbers and convert them to integers
                row = [int(num) for num in line.split()]
                array.append(row)
    return array


grid = read()


def createChild(size):
    global grid

    children = []

    for i in range(size):
        child = [0] * 81
        for row in range(9):
            index = 0
            randomArray = [1, 2, 3, 4, 5, 6, 7, 8, 9]
            random.shuffle(randomArray)
            for coulm in range(9):
                if grid[row][coulm] != 0:
                    child[row * 9 + coulm] = grid[row][coulm]
                else:
                    while randomArray[index] in child[9 * row:9 * (row + 1)] or randomArray[index] in grid[row]:
                        index += 1
                    child[row * 9 + coulm] = randomArray[index]
                    tmp = child.copy()

        children.append(child)
    return children


def crossover_func(parents, offspring_size, ga_instance):
    offspring = []


    if ga_instance.generations_completed == 0:
        return numpy.array(createChild(offspring_size[0]))

    idx = 0
    while len(offspring) != offspring_size[0]:
        parent1 = parents[idx % parents.shape[0], :].copy()
        parent2 = parents[(idx + 1) % parents.shape[0], :].copy()

        parent1NRows = numpy.random.choice([1, 2, 3, 4, 5, 6, 7, 8])
        randomArray = [1, 2, 3, 4, 5, 6, 7, 8, 0]
        random.shuffle(randomArray)
        randomArray = randomArray[:parent1NRows]

        for i in randomArray:
            parent2[i * 9:(i + 1) * 9] = parent1[i * 9:(i + 1) * 9]

        offspring.append(parent2)

        idx += 1

    return numpy.array(offspring)


def mutation_func(offspring, ga_instance):
    # za svako rješenje u generacijiz
    for instance in offspring:

        # na svako rješenje primjeni mutaciju numberOfMutations puta
        for i in range(numberOfMutations):
            # odaberi random redak
            randomRow = random.randint(0, 8)

            # odaberi random stupac koji nije fiksiran
            randomNumbFrom = random.randint(0, 8)
            while grid[randomRow][randomNumbFrom] != 0:
                randomNumbFrom = random.randint(0, 8)

            # odaberi random stupac koji nije fiksiran
            randomNumbTo = random.randint(0, 8)
            while grid[randomRow][randomNumbTo] != 0:
                randomNumbTo = random.randint(0, 8)

            # zamijeni brojeve na odabranim pozicijama
            tmp = instance[randomRow * 9 + randomNumbFrom]
            instance[randomRow * 9 + randomNumbFrom] = instance[randomRow * 9 + randomNumbTo]
            instance[randomRow * 9 + randomNumbTo] = tmp

    return offspring


def fitness_func(ga_instance, solution, solution_idx):
    grid = solution.reshape((9, 9))
    boxovi = [[0] * 9 for _ in range(9)]

    fitness = 0
    unique = 0
    good = 0
    row = [0] * 9

    for i in range(81):
        number = grid[i // 9][i % 9]

        if row[number - 1] == 0:
            row[number - 1] = 1
            unique += 1
        else:
            unique -= 1

        boxNumber = (i // 27) * 3 + (i % 9) // 3

        if boxovi[boxNumber][number - 1] == 0:
            boxovi[boxNumber][number - 1] = 1

        if i % 9 == 8:
            if unique == 9:
                good += 1
            fitness += unique
            unique = 0
            row = [0] * 9  # Reset the row array

    colum = [0] * 9

    for i in range(81):

        number = grid[i % 9][i // 9]
        if colum[number - 1] == 0:
            colum[number - 1] = 1
            unique += 1
        else:
            unique -= 1

        if i % 9 == 8:
            if unique == 9:
                good += 1

            fitness += unique
            unique = 0
            colum = [0] * 9  # Reset the row array

    for box in boxovi:
        for number in box:
            if number == 1:
                unique += 1
            else:
                unique -= 1

        if unique == 9:
            good += 1
        fitness += unique
        unique = 0

        global genBest
        if solution_idx == 0:
            genBest = fitness
        elif fitness > genBest:
            genBest = fitness

    return fitness


# funkcija koja opisuje ispis za svaku pojedinu generaciju
def on_generation(ga_instance):
    fitness_values.append(ga_instance.best_solution()[1])
    print("Generation : ", ga_instance.generations_completed)
    print("Fitness of the best solution :", ga_instance.best_solution()[1])

    # best = ga_instance.best_solution()[0].reshape(9, 9)
    # for row in best:
    #   print(row)


# konfiguracija nekih parametara GA
num_generations = 10000  # broj generacija
num_parents_mating = 3  # broj rješenja koja su odabrana kao roditelji sljedeće generacije
sol_per_pop = 20  # broj rješenja u jednoj generaciji
num_genes = 81  # broj gena u jednom rješenju/kromosomu
gene_space = [1, 2, 3, 4, 5, 6, 7, 8, 9]

# zadavanje parametara PyGadu - ostale parametre naći na linku gore
ga_instance = pygad.GA(num_generations=num_generations,
                       num_parents_mating=num_parents_mating,
                       sol_per_pop=sol_per_pop,
                       num_genes=num_genes,
                       gene_space=gene_space,
                       fitness_func=fitness_func,
                       mutation_type=mutation_func,
                       on_generation=on_generation,
                       gene_type=int,
                       crossover_type=crossover_func,
                       stop_criteria=["reach_243"],
                       )  # ostavi 3 najbolja rješenja iz prethodne generacije

# pokretanje
ga_instance.run()

# detalji najboljeg rješenja
solution, solution_fitness, solution_idx = ga_instance.best_solution(ga_instance.last_generation_fitness)
print("Solved sudoku")


def printSudoku(solution):
    for i in range(9):
        for j in range(9):
            print(solution[i * 9 + j], end=" ")
            if j % 3 == 2:
                print("|", end=" ")
        print()
        if i % 3 == 2:
            print("-----------------------")

    pass


printSudoku(solution)
print(f"Fitness value of the best solution = {solution_fitness}")
print(f"Index of the best solution : {solution_idx}")
if ga_instance.best_solution_generation != -1:
    print(f"Best fitness value reached after {ga_instance.best_solution_generation} generations.")
    
# plot fitnesa
plt.figure()
plt.plot(fitness_values)
plt.title("Fitness Over Generations")
plt.xlabel("Generation")
plt.ylabel("Fitness Value")
plt.savefig("./images/Sudoku.png")
plt.close()
