import pygad
import numpy as np
import matplotlib.pylab as plt

# List to store the fitness outputs for each generation
fitness_outputs = []

# List of objects with their dimensions
objects = [
    (10, 5), (15, 8), (7, 12), (20, 10), (6, 15),
    (12, 8), (18, 6), (10, 10), (8, 12), (14, 7),
    (16, 9), (9, 14), (11, 10), (13, 8), (17, 7),
    (8, 15), (10, 13), (15, 10), (12, 12), (9, 18),
    (20, 6), (14, 9), (7, 14), (16, 7), (10, 10),
    (12, 10), (18, 5), (15, 9), (9, 12), (10, 14),
    (12, 7), (11, 12), (13, 10), (8, 16), (10, 9),
    (15, 7), (14, 11), (7, 16), (13, 9), (12, 11),
    (9, 15), (10, 12), (11, 8), (14, 10), (8, 14),
    (16, 6), (13, 11), (10, 11), (12, 8), (17, 6)
]

# Maximum dimensions of the sheet
maxWidth = 100
maxHeight = 50

# Function to calculate fitness of a solution
def fitness_func(ga_instance, solution, solution_idx):
    total_used_area = 0
    total_unused_area = maxWidth * maxHeight

    # Calculate the total used area by placing each object on the sheet
    for i, count in enumerate(solution):
        total_used_area += count * objects[i][0] * objects[i][1]

    # If the used area exceeds the total available area, penalize the solution
    if total_used_area > total_unused_area:
        return total_unused_area - total_used_area
    else:
        return total_used_area

# Callback function called on each generation
def on_generation(ga_instance):
    print("Generation : ", ga_instance.generations_completed)
    print("Fitness of the best solution :", ga_instance.best_solution()[1])
    
    # Calculate waste percentage and append to the fitness_outputs list
    fitness_outputs.append(round(1 - (ga_instance.best_solution()[1] / (maxHeight * maxWidth)), 3) * 100)

# Number of generations, population size, and number of genes
num_generations = 100
sol_per_pop = 100
num_genes = len(objects)

# Gene space (0 or 1) for binary representation of solutions
gene_space = [0, 1]

# Initialize the population with all zeros
initial_population = np.zeros((sol_per_pop, num_genes), dtype=int)

# Create the genetic algorithm instance
ga_instance = pygad.GA(num_generations=num_generations,
                      num_parents_mating=3,
                      sol_per_pop=sol_per_pop,
                      num_genes=num_genes,
                      gene_space=gene_space,
                      fitness_func=fitness_func,
                      on_generation=on_generation,
                      mutation_num_genes=10,
                      initial_population=initial_population,
                      parent_selection_type="tournament",
                      K_tournament=3,
                      )

# Run the genetic algorithm
ga_instance.run()

# Get the best solution from the last generation
solution, solution_fitness, solution_idx = ga_instance.best_solution(ga_instance.last_generation_fitness)
print(f"Parameters of the best solution : {solution}")

# Plotting the results using Matplotlib
indices = np.arange(len(fitness_outputs))
max_index = fitness_outputs.index(max(fitness_outputs))
min_index = fitness_outputs.index(min(fitness_outputs))

plt.plot(indices, fitness_outputs, marker='o', linestyle='-', color='blue', label='Waste Percentage')

plt.scatter(min_index, min(fitness_outputs), color='red', marker='o', s=100, label=f'Best Solution: {min(fitness_outputs)}% of waste')

plt.title('Graph of waste from a sheet of material')
plt.xlabel('Generation number')
plt.ylabel('% of waste area')

plt.legend()

# Show the plot
plt.savefig("./images/rezanje.png")
