import pygad
import numpy
import matplotlib.pyplot as plt

"""
PRIKAZ KORIŠTENJA PYGADA NA PRIMJERU PROBLEMA RUKSAKA

instalacija PyGada i detaljnije upute: https://pygad.readthedocs.io/en/latest/index.html
"""
fitness_values = []

# ulazni podaci - vrijednosti, težine i nosivost ruksaka
values = [1, 2, 3, 4, 7, 8, 9, 10, 11, 15, 16, 17, 21, 22, 28, 30, 32, 34, 35, 37, 38, 41, 43, 48, 49, 50, 51,
          52, 53, 54, 55, 56, 57, 59, 61, 64, 67, 70, 71, 73, 74, 79, 82, 84, 85, 89, 90, 92, 96, 100]  # Function inputs.
weights = [1, 2, 3, 4, 52, 54, 56, 59, 60, 61, 62, 72, 77, 79, 80, 81, 83, 84, 87, 63, 64, 66, 67, 68,
           70, 71,  88, 8, 9, 11, 12, 16, 21, 22, 23, 28, 29, 34, 36, 40, 41, 42, 47, 48, 90, 91, 92, 93, 96, 100]
maxWeight = 500  # Function output.

# fitnes funkcija za izračun fitnesa jednog člana


def fitness_func(ga_instance, solution, solution_idx):
    # računanje vrijednosti svih predmeta uključenih u ruksak
    value = numpy.sum(numpy.multiply(solution, values))
    # računanje težina svih predmeta uključenih u ruksak
    weight = numpy.sum(numpy.multiply(solution, weights))
    if weight > maxWeight:					# ako je težina predmeta veća od nosivosti fitnes pokazuje koliko je premašena masa
        return maxWeight-weight
    else: 							# inače se vraća vrijednost predmeta u ruksaku
        return value

# funkcija koja opisuje ispis za svaku pojedinu generaciju


def update_fitness_plot(ga_instance):
    fitness_values.append(ga_instance.best_solution()[1])
    print("Generation : ", ga_instance.generations_completed)
    print("Fitness of the best solution :", ga_instance.best_solution()[1])



# konfiguracija nekih parametara GA
num_generations = 1000  # broj generacija
# broj rješenja koja su odabrana kao roditelji sljedeće generacije
num_parents_mating = 10
sol_per_pop = 100  # broj rješenja u jednoj generaciji
num_genes = 50  # broj gena u jednom rješenju/kromosomu
# mogući geni (u ovom slučaju 0 ili 1 - predmet nije/je uključen u ruksak)
gene_space = [0, 1]

# zadavanje parametara PyGadu - ostale parametre naći na linku gore
ga_instance = pygad.GA(
    num_generations=num_generations,
    num_parents_mating=num_parents_mating,
    sol_per_pop=sol_per_pop,
    num_genes=num_genes,
    gene_space=gene_space,
    fitness_func=fitness_func,
    on_generation=lambda ga_instance: update_fitness_plot(ga_instance),
    mutation_num_genes=10,
    random_mutation_min_val=0,
    random_mutation_max_val=0,
    keep_elitism=3)

# pokretanje
ga_instance.run()

# detalji najboljeg rješenja
solution, solution_fitness, solution_idx = ga_instance.best_solution(
    ga_instance.last_generation_fitness)
print(f"Parameters of the best solution : {solution}")
print(f"Fitness value of the best solution = {solution_fitness}")
print(f"Index of the best solution : {solution_idx}")
print(solution)
if ga_instance.best_solution_generation != -1:
    print(
        f"Best fitness value reached after {ga_instance.best_solution_generation} generations.")
        
        
# Plot and save the fitness values
plt.figure()
plt.plot(fitness_values)
plt.title("Fitness Over Generations")
plt.xlabel("Generation")
plt.ylabel("Fitness Value")
plt.savefig("./images/ruksak.png")
plt.close()
